=== modified file 'CMakeLists.txt'
--- CMakeLists.txt	2016-06-14 08:33:42 +0000
+++ CMakeLists.txt	2016-07-01 16:15:54 +0000
@@ -84,8 +84,7 @@
 pkg_check_modules(QTDBUSTEST libqtdbustest-1 REQUIRED)
 pkg_check_modules(QTDBUSMOCK libqtdbusmock-1 REQUIRED)
 pkg_check_modules(APPLICATION_API REQUIRED unity-shell-application=18)
-
-include_directories(${APPLICATION_API_INCLUDE_DIRS})
+pkg_check_modules(CGMANAGER libcgmanager REQUIRED)
 
 add_definitions(-DMIR_REQUIRE_DEPRECATED_EVENT_OPT_IN=1)
 

=== modified file 'debian/control'
--- debian/control	2016-06-03 11:16:30 +0000
+++ debian/control	2016-07-01 16:15:54 +0000
@@ -6,6 +6,7 @@
                cmake-extras (>= 0.3),
                debhelper (>= 9),
                google-mock (>= 1.6.0+svn437),
+               libcgmanager-dev,
                libfontconfig1-dev,
                libgles2-mesa-dev,
                libglib2.0-dev,

=== modified file 'src/modules/Unity/Application/CMakeLists.txt'
--- src/modules/Unity/Application/CMakeLists.txt	2016-06-23 19:58:48 +0000
+++ src/modules/Unity/Application/CMakeLists.txt	2016-07-01 16:15:54 +0000
@@ -10,6 +10,7 @@
     ${CMAKE_SOURCE_DIR}/src/platforms/mirserver
     ${CMAKE_SOURCE_DIR}/src/common
     ${GSETTINGS_QT_INCLUDE_DIRS}
+    ${CGMANAGER_INCLUDE_DIRS}
 
     ${LTTNG_INCLUDE_DIRS}
     ${Qt5Gui_PRIVATE_INCLUDE_DIRS}
@@ -25,6 +26,7 @@
 set(QMLMIRPLUGIN_SRC
     application_manager.cpp
     application.cpp
+    cgmanager.cpp
     ../../../common/abstractdbusservicemonitor.cpp
     ../../../common/debughelpers.cpp
     dbusfocusinfo.cpp

=== added file 'src/modules/Unity/Application/cgmanager.cpp'
--- src/modules/Unity/Application/cgmanager.cpp	1970-01-01 00:00:00 +0000
+++ src/modules/Unity/Application/cgmanager.cpp	2016-07-01 16:15:54 +0000
@@ -0,0 +1,101 @@
+#include "cgmanager.h"
+
+// QPA mirserver
+#include <logging.h>
+
+#include <cgmanager/cgmanager.h>
+
+#include <QDBusArgument>
+#include <QDBusMessage>
+
+using namespace qtmir;
+
+// According to D-Bus interface defined in:
+// https://github.com/lxc/cgmanager/blob/master/org.linuxcontainers.cgmanager.xml
+
+CGManager::CGManager(QObject *parent)
+    : QObject(parent)
+{
+}
+
+CGManager::~CGManager()
+{
+    QDBusConnection::disconnectFromPeer("cgmanager");
+}
+
+QDBusConnection CGManager::getConnection()
+{
+    auto connection = QDBusConnection("cgmanager");
+
+    if (!connection.isConnected()) {
+        connection = QDBusConnection::connectToPeer(CGMANAGER_DBUS_PATH, "cgmanager");
+        if (!connection.isConnected()) {
+            qCWarning(QTMIR_DBUS, "CGManager: Failed to connect to %s", CGMANAGER_DBUS_PATH);
+        }
+    }
+
+    return connection;
+}
+
+QString CGManager::getCGroupOfPid(const QString &controller, pid_t pid)
+{
+    auto connection = getConnection();
+    if (!connection.isConnected()) {
+        return QString();
+    }
+
+    auto message = QDBusMessage::createMethodCall(QString() /*service*/, m_path, m_interface, "GetPidCgroup");
+
+    QList<QVariant> arguments;
+    arguments << QVariant(controller);
+    arguments << QVariant((int)pid);
+
+    message.setArguments(arguments);
+
+    QDBusMessage reply = connection.call(message);
+
+    if (reply.type() == QDBusMessage::ReplyMessage) {
+        return reply.arguments()[0].toString();
+    } else {
+        Q_ASSERT(reply.type() == QDBusMessage::ErrorMessage);
+        qCWarning(QTMIR_DBUS) << "CGManager::getCGroupOfPid failed." << reply.errorMessage();
+        return QString();
+    }
+}
+
+QSet<pid_t> CGManager::getTasks(const QString &controller, const QString &cgroup)
+{
+    auto connection = getConnection();
+    if (!connection.isConnected()) {
+        return QSet<pid_t>();
+    }
+
+    auto message = QDBusMessage::createMethodCall(QString() /*service*/, m_path, m_interface, "GetTasks");
+
+    QList<QVariant> arguments;
+    arguments << QVariant(controller);
+    arguments << QVariant(cgroup);
+
+    message.setArguments(arguments);
+
+    QDBusMessage reply = connection.call(message);
+
+    if (reply.type() == QDBusMessage::ErrorMessage) {
+        qCWarning(QTMIR_DBUS) << "CGManager::getCGroupOfPid failed." << reply.errorMessage();
+        return QSet<pid_t>();
+    }
+
+    Q_ASSERT(reply.type() == QDBusMessage::ReplyMessage);
+
+    auto arg = reply.arguments()[0].value<QDBusArgument>();
+    Q_ASSERT(arg.currentType() == QDBusArgument::ArrayType);
+
+    auto pidList = qdbus_cast<QList<int>>(arg);
+    Q_ASSERT(pidList.count() > 0);
+
+    QSet<pid_t> pidSet;
+    for (int i = 0; i < pidList.count(); ++i) {
+        pidSet << (pid_t)pidList[i];
+    }
+    return pidSet;
+}

=== added file 'src/modules/Unity/Application/cgmanager.h'
--- src/modules/Unity/Application/cgmanager.h	1970-01-01 00:00:00 +0000
+++ src/modules/Unity/Application/cgmanager.h	2016-07-01 16:15:54 +0000
@@ -0,0 +1,42 @@
+/*
+ * Copyright (C) 2016 Canonical, Ltd.
+ *
+ * This program is free software: you can redistribute it and/or modify it under
+ * the terms of the GNU Lesser General Public License version 3, as published by
+ * the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
+ * SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ */
+
+#ifndef QTMIR_CGMANAGER_H
+#define QTMIR_CGMANAGER_H
+
+#include <QDBusConnection>
+
+namespace qtmir {
+
+class CGManager : public QObject {
+    Q_OBJECT
+public:
+    CGManager(QObject *parent = nullptr);
+    virtual ~CGManager();
+
+    QString getCGroupOfPid(const QString &controller, pid_t pid);
+
+    QSet<pid_t> getTasks(const QString &controller, const QString &cgroup);
+
+private:
+    const QString m_interface{"org.linuxcontainers.cgmanager0_0"};
+    const QString m_path{"/org/linuxcontainers/cgmanager"};
+    QDBusConnection getConnection();
+};
+
+} // namespace qtmir
+
+#endif // QTMIR_CGMANAGER_H

=== modified file 'src/modules/Unity/Application/dbusfocusinfo.cpp'
--- src/modules/Unity/Application/dbusfocusinfo.cpp	2016-06-24 16:28:37 +0000
+++ src/modules/Unity/Application/dbusfocusinfo.cpp	2016-07-01 16:15:54 +0000
@@ -16,6 +16,12 @@
 
 #include "dbusfocusinfo.h"
 
+// local
+#include "cgmanager.h"
+
+// QPA mirserver
+#include <logging.h>
+
 #include <QDBusConnection>
 
 using namespace qtmir;
@@ -25,6 +31,8 @@
 {
     QDBusConnection::sessionBus().registerService("com.canonical.Unity.FocusInfo");
     QDBusConnection::sessionBus().registerObject("/", this, QDBusConnection::ExportScriptableSlots);
+
+    m_cgManager = new CGManager(this);
 }
 
 bool DBusFocusInfo::isPidFocused(unsigned int pid)
@@ -34,22 +42,42 @@
         // Don't bother checking if it has a QML with activeFocus() which is not a MirSurfaceItem.
         return true;
     } else {
-        SessionInterface *session = findSessionWithPid(pid);
+        auto pidSet = fetchAssociatedPids((pid_t)pid);
+        SessionInterface *session = findSessionWithPid(pidSet);
         return session ? session->activeFocus() : false;
     }
 }
 
-SessionInterface* DBusFocusInfo::findSessionWithPid(unsigned int uintPid)
-{
-    pid_t pid = (pid_t)uintPid;
+QSet<pid_t> DBusFocusInfo::fetchAssociatedPids(pid_t pid)
+{
+    QString cgroup = m_cgManager->getCGroupOfPid("freezer", pid);
+
+    // If a cgroup has a format like this:
+    // /user.slice/user-32011.slice/session-c3.scope/upstart/application-legacy-puritine_gedit_0.0-
+    // All PIds in it are associated with a single application.
+    if (cgroup.split("/").contains("upstart")) {
+        QSet<pid_t> pidSet = m_cgManager->getTasks("freezer", cgroup);
+        qCDebug(QTMIR_DBUS) << "DBusFocusInfo: pid" << pid << "is in cgroup" << cgroup << "along with:" << pidSet;
+        if (pidSet.isEmpty()) {
+            pidSet << pid;
+        }
+        return pidSet;
+    } else {
+        qCDebug(QTMIR_DBUS) << "DBusFocusInfo: pid" << pid << "is in cgroup" << cgroup << "which is not app-specific";
+        return QSet<pid_t>({pid});
+    }
+}
+
+SessionInterface* DBusFocusInfo::findSessionWithPid(const QSet<pid_t> &pidSet)
+{
     Q_FOREACH (Application* application, m_applications) {
         auto session = application->session();
-        if (session->pid() == pid) {
+        if (pidSet.contains(session->pid())) {
             return session;
         }
         SessionInterface *chosenChildSession = nullptr;
         session->foreachChildSession([&](SessionInterface* childSession) {
-            if (childSession->pid() == pid) {
+            if (pidSet.contains(childSession->pid())) {
                 chosenChildSession = childSession;
             }
         });

=== modified file 'src/modules/Unity/Application/dbusfocusinfo.h'
--- src/modules/Unity/Application/dbusfocusinfo.h	2016-06-24 16:28:37 +0000
+++ src/modules/Unity/Application/dbusfocusinfo.h	2016-07-01 16:15:54 +0000
@@ -15,11 +15,14 @@
  */
 
 #include <QList>
+#include <QSet>
 
 #include "application.h"
 
 namespace qtmir {
 
+class CGManager;
+
 /*
   FIXME: This is a hack to provide OSK with needed info for avoiding input snooping.
          Remove when possible
@@ -40,9 +43,12 @@
     Q_SCRIPTABLE bool isPidFocused(unsigned int pid);
 
 private:
-    SessionInterface* findSessionWithPid(unsigned int uintPid);
+    QSet<pid_t> fetchAssociatedPids(pid_t pid);
+    SessionInterface* findSessionWithPid(const QSet<pid_t> &pidSet);
 
     const QList<Application*> &m_applications;
+
+    CGManager *m_cgManager;
 };
 
 } // namespace qtmir

=== modified file 'src/platforms/mirserver/logging.cpp'
--- src/platforms/mirserver/logging.cpp	2016-05-03 19:45:51 +0000
+++ src/platforms/mirserver/logging.cpp	2016-07-01 16:15:54 +0000
@@ -24,3 +24,4 @@
 Q_LOGGING_CATEGORY(QTMIR_CLIPBOARD, "qtmir.clipboard")
 Q_LOGGING_CATEGORY(QTMIR_SENSOR_MESSAGES, "qtmir.sensor")
 Q_LOGGING_CATEGORY(QTMIR_SCREENS, "qtmir.screens")
+Q_LOGGING_CATEGORY(QTMIR_DBUS, "qtmir.dbus", QtWarningMsg)

=== modified file 'src/platforms/mirserver/logging.h'
--- src/platforms/mirserver/logging.h	2015-08-20 10:16:54 +0000
+++ src/platforms/mirserver/logging.h	2016-07-01 16:15:54 +0000
@@ -26,5 +26,6 @@
 Q_DECLARE_LOGGING_CATEGORY(QTMIR_MIR_INPUT)
 Q_DECLARE_LOGGING_CATEGORY(QTMIR_CLIPBOARD)
 Q_DECLARE_LOGGING_CATEGORY(QTMIR_SCREENS)
+Q_DECLARE_LOGGING_CATEGORY(QTMIR_DBUS)
 
 #endif // UBUNTU_APPLICATION_PLUGIN_LOGGING_H

